module DynamicArchive
  class AssocOptions
    attr_accessor(
      :foreign_key,
      :class_name,
      :primary_key
    )

    def model_class
      begin
        class_name.constantize
      rescue NameError
        Object.const_set(class_name, Class.new(DynamicArchive::Base))
      ensure
        class_name.constantize
      end
    end

    def table_name
      model_class.table_name
    end
  end

  class BelongsToOptions < DynamicArchive::AssocOptions
    def initialize(name, options = {})
      defaults = {
        foreign_key: "#{name.to_s.underscore}_id".to_sym,
        primary_key: :id,
        class_name:  name.to_s.camelcase
      }
      options = defaults.merge(options)

      @foreign_key = options[:foreign_key]
      @primary_key = options[:primary_key]
      @class_name  = options[:class_name]
    end
  end

  class HasManyOptions < DynamicArchive::AssocOptions
    def initialize(name, self_class_name, options = {})
      defaults = {
        foreign_key: "#{self_class_name.to_s.underscore}_id".to_sym,
        primary_key: :id,
        class_name: name.to_s.singularize.camelcase
      }
      options = defaults.merge(options)

      @foreign_key = options[:foreign_key]
      @primary_key = options[:primary_key]
      @class_name  = options[:class_name]
    end
  end

  module Associatable
    def assoc_options
      @assoc_options ||= {}
    end

    def belongs_to(name, options = {})
      options = BelongsToOptions.new(name, options)
      assoc_options[name] = options

      define_method(name) do
        foreign_key_val = self.send(options.foreign_key)

        options.model_class.where(options.primary_key => foreign_key_val).first
      end

    end

    def has_many(name, options = {})
      options = HasManyOptions.new(name, self.to_s, options)
      assoc_options[name] = options

      define_method(name) do
        foreign_key_val = self.send(options.primary_key)

        options.model_class.where(options.foreign_key => foreign_key_val)
      end
    end

    def has_one_through(name, through_name, source_name)
      through_options = self.assoc_options[through_name]
      source_options = through_options.model_class.assoc_options[source_name]

      define_method(name) do
        key = send(through_options.foreign_key)
        # use `LIKE` to workaround SQLite bug
        results = DBConnection.execute(<<-SQL, key)
          SELECT
            #{source_options.table_name}.*
          FROM
            #{through_options.table_name}
          JOIN
            #{source_options.table_name}
          ON
            #{through_options.table_name}.#{source_options.foreign_key} LIKE
            #{source_options.table_name}.#{source_options.primary_key}
          WHERE
            #{through_options.table_name}.#{through_options.primary_key} LIKE ?
        SQL

        source_options.model_class.parse_all(results).first
      end
    end

    def has_many_through(name, through_name, source_name)
      through_options = self.assoc_options[through_name]
      source_options = through_options.model_class.assoc_options[source_name]

      define_method(name) do
        if source_options.is_a?(HasManyOptions)
          key = send(through_options.primary_key)
          # use `LIKE` to workaround SQLite bug
          results = DBConnection.execute(<<-SQL, key)
            SELECT
              #{source_options.table_name}.*
            FROM
              #{through_options.table_name}
            JOIN
              #{source_options.table_name}
            ON
              #{through_options.table_name}.#{source_options.primary_key} LIKE
              #{source_options.table_name}.#{source_options.foreign_key}
            WHERE
              #{through_options.table_name}.#{through_options.foreign_key} LIKE ?
          SQL

        else
          key = send(through_options.primary_key)
          # use `LIKE` to workaround SQLite bug
          results = DBConnection.execute(<<-SQL, key)
            SELECT
              #{source_options.table_name}.*
            FROM
              #{through_options.table_name}
            JOIN
              #{source_options.table_name}
            ON
              #{through_options.table_name}.#{source_options.foreign_key} LIKE
              #{source_options.table_name}.#{source_options.primary_key}
            WHERE
              #{through_options.table_name}.#{through_options.foreign_key} LIKE ?
          SQL
        end

        source_options.model_class.parse_all(results)
      end
    end
  end
end
