require_relative 'db_connection'
require_relative 'associatable'
require_relative 'searchable'
require_relative 'validatable'
require 'active_support/inflector'

module DynamicArchive
  class Base
    extend DynamicArchive::Searchable
    extend DynamicArchive::Associatable
    extend DynamicArchive::Validatable

    def self.columns
      @columns ||= DBConnection.execute2(<<-SQL).first.map(&:to_sym)
        SELECT
          *
        FROM
          #{table_name}
      SQL
    end

    def self.finalize!
      columns.each do |column|
        define_method("#{column}=") do |message|
          attributes[column] = message
        end

        define_method("#{column}") do
          attributes[column]
        end
      end
    end

    def self.table_name=(table_name)
      @table_name = table_name
    end

    def self.table_name
      @table_name ||= self.to_s.tableize
    end

    def self.all
      results = DBConnection.execute(<<-SQL)
        SELECT
          *
        FROM
          #{table_name}
      SQL
      parse_all(results)
    end

    def self.parse_all(results)
      results.map do |result|
        params = {}
        result.each do |key, value|
          params[key.to_sym] = value
        end
        self.new(params)
      end
    end

    def self.find(id)
      results = DBConnection.execute(<<-SQL, id.to_i)
        SELECT
          *
        FROM
          #{table_name}
        WHERE
          id = (?)
      SQL
      parse_all(results).first
    end

    def initialize(params = {})
      @validations = Hash.new { |h, k| h[k] = {} }

      params.each do |key, val|
        attr_name = key.to_sym
        unless self.class.columns.include?(key) || self.class.method_defined?("#{attr_name}=")
          raise "unknown attribute '#{attr_name}'"
        end
        self.send("#{attr_name}=", val)
      end
    end

    def attributes
      @attributes ||= {}
    end

    def attribute_values
      self.class.columns.map do |attr|
        self.send(attr)
      end
    end

    def insert
      col_names = self.class.columns.join(", ")
      question_marks = (["?"] * self.class.columns.length).join(", ")
      DBConnection.execute(<<-SQL, attribute_values)
        INSERT INTO
          #{self.class.table_name} (#{col_names})
        VALUES
          (#{question_marks})
      SQL
      self.id = DBConnection.last_insert_row_id
    end

    def update
      set_line = self.class.columns.map { |col| "#{col} = (?)"}.join(", ")
      DBConnection.execute(<<-SQL, attribute_values)
        UPDATE
          #{self.class.table_name}
        SET
          #{set_line}
        WHERE
          id = #{id}
      SQL
    end

    def save
      if self.class.is_valid?(self)
        id.nil? ? insert : update
      else
        false
      end
    end

    def destroy
      DBConnection.execute(<<-SQL, self.id)
        DELETE FROM
          #{self.class.table_name}
        WHERE
          id = (?)
      SQL

      self
    end

    def errors
      @errors ||= []
    end
  end
end
