require 'sqlite3'

PRINT_QUERIES = ENV['PRINT_QUERIES'] == 'true'
MIGRATION_FILES = Dir.glob("#{ENV['MIGRATIONS_FOLDER']}/*.sql").to_a

class DBConnection
  def self.open(db_file_name)
    options = {
      results_as_hash: true,
      type_translation: true
    }
    @db = SQLite3::Database.new(db_file_name, options)

    @db
  end

  def self.db
    @db
  end

  def self.reset
    `rm '#{ENV['DB_FILE']}'`
    `rm '#{ENV['SCHEMA_FILE']}'`
    self.open(ENV['DB_FILE'])

    self.migrate
    self.seed
  end

  def self.instance
    self.open(ENV['DB_FILE']) if @db.nil?

    @db
  end

  def self.execute(*args)
    print_query(*args)
    instance.execute(*args)
  end

  def self.execute2(*args)
    print_query(*args)
    instance.execute2(*args)
  end

  def self.execute_batch(statement)
    print_query(statement)
    instance.execute_batch(statement)
  end

  def self.last_insert_row_id
    instance.last_insert_row_id
  end

  def self.migrate(resetting = false)
    self.ensure_migrations_table
    already_migrated = self.previous_migrations

    MIGRATION_FILES.each do |migration_file|
      migration_file_name = (/\/(\w+)\.sql$/).match(migration_file)[1]
      next if already_migrated.include?(migration_file_name)

      migration = IO.read(migration_file)
      self.execute(migration)
      `echo '#{migration}' >> '#{ENV['SCHEMA_FILE']}'`
      self.record_migration(migration_file_name)
    end

    @db
  end

  def self.reset_schema
    `rm '#{ENV['SCHEMA_FILE']}'`
    self.execute(<<-SQL)
      DROP TABLE IF EXISTS migrations
    SQL
    `touch '#{ENV['SCHEMA_FILE']}'`
  end

  def self.seed
    seed_data = IO.read(ENV['SEED_FILE'])
    self.execute_batch(seed_data)
  end

  private

  def self.record_migration(file_name)
    self.execute(<<-SQL, file_name)
      INSERT INTO
        migrations (file_name)
      VALUES
        (?)
    SQL
  end

  def self.ensure_migrations_table
    self.execute(<<-SQL)
        CREATE TABLE IF NOT EXISTS migrations (
          id INTEGER PRIMARY KEY,
          file_name VARCHAR(255) NOT NULL
        );
    SQL
  end

  def self.previous_migrations
  ensure_migrations_table
    migrations = self.execute(<<-SQL)
        SELECT
          file_name
        FROM
          migrations
    SQL

    migrations.map do |migration|
      migration["file_name"]
    end
  end

  def self.create_table(table_name, columns = {})
    column_text = columns.map { |name, options| "#{name} #{options}" }.join(", ")
    self.execute(<<-SQL)
      CREATE TABLE #{table_name} (
        id INTEGER PRIMARY KEY,
        #{column_text}
      );
    SQL
  end

  def self.add_column(table_name, column_name, column_type)
    self.execute(<<-SQL)
      ALTER TABLE #{table_name} ADD #{column_name} #{column_type};
    SQL
  end

  def self.remove_column(table_name, column_name)
    self.execute(<<-SQL)
      ALTER TABLE #{table_name} DROP COLUMN #{column_name};
    SQL
  end

  def self.print_query(query, *interpolation_args)
    return unless PRINT_QUERIES

    puts '--------------------'
    puts query
    unless interpolation_args.empty?
      puts "interpolate: #{interpolation_args.inspect}"
    end
    puts '--------------------'
  end
end
