require_relative '../testing_database/database'
require 'associatable'

describe 'AssocOptions' do
  describe 'DynamicArchive::BelongsToOptions' do
    it 'provides defaults' do
      options = DynamicArchive::BelongsToOptions.new('house')

      expect(options.foreign_key).to eq(:house_id)
      expect(options.class_name).to eq('House')
      expect(options.primary_key).to eq(:id)
    end

    it 'allows overrides' do
      options = DynamicArchive::BelongsToOptions.new('owner',
                                     foreign_key: :human_id,
                                     class_name: 'Human',
                                     primary_key: :human_id
      )

      expect(options.foreign_key).to eq(:human_id)
      expect(options.class_name).to eq('Human')
      expect(options.primary_key).to eq(:human_id)
    end
  end

  describe 'DynamicArchive::HasManyOptions' do
    it 'provides defaults' do
      options = DynamicArchive::HasManyOptions.new('cats', 'Human')

      expect(options.foreign_key).to eq(:human_id)
      expect(options.class_name).to eq('Cat')
      expect(options.primary_key).to eq(:id)
    end

    it 'allows overrides' do
      options = DynamicArchive::HasManyOptions.new('cats', 'Human',
                                   foreign_key: :owner_id,
                                   class_name: 'Kitten',
                                   primary_key: :human_id
      )

      expect(options.foreign_key).to eq(:owner_id)
      expect(options.class_name).to eq('Kitten')
      expect(options.primary_key).to eq(:human_id)
    end
  end

  describe 'AssocOptions' do
    before(:all) do
      class Cat < DynamicArchive::Base
        self.finalize!
      end

      class Human < DynamicArchive::Base
        self.table_name = 'humans'

        self.finalize!
      end
    end

    it '#model_class returns class of associated object' do
      options = DynamicArchive::BelongsToOptions.new('human')
      expect(options.model_class).to eq(Human)

      options = DynamicArchive::HasManyOptions.new('cats', 'Human')
      expect(options.model_class).to eq(Cat)
    end

    it '#table_name returns table name of associated object' do
      options = DynamicArchive::BelongsToOptions.new('human')
      expect(options.table_name).to eq('humans')

      options = DynamicArchive::HasManyOptions.new('cats', 'Human')
      expect(options.table_name).to eq('cats')
    end
  end
end

describe 'Associatable' do
  before(:each) { DBConnection.reset }
  after(:each) { DBConnection.reset }

  before(:all) do
    class Cat < DynamicArchive::Base
      belongs_to :human, foreign_key: :owner_id
      has_many :play_times

      finalize!
    end

    class Human < DynamicArchive::Base
      self.table_name = 'humans'

      has_many :cats, foreign_key: :owner_id
      belongs_to :house

      finalize!
    end

    class House < DynamicArchive::Base
      has_many :humans

      finalize!
    end

    class Toy < DynamicArchive::Base
      has_many :play_times
    end

    class PlayTime < DynamicArchive::Base
      belongs_to :toy
      belongs_to :cat
    end
  end

  describe '#belongs_to' do
    let(:breakfast) { Cat.find(1) }
    let(:devon) { Human.find(1) }

    it 'fetches `human` from `Cat` correctly' do
      expect(breakfast).to respond_to(:human)
      human = breakfast.human

      expect(human).to be_instance_of(Human)
      expect(human.fname).to eq('Devon')
    end

    it 'fetches `house` from `Human` correctly' do
      expect(devon).to respond_to(:house)
      house = devon.house

      expect(house).to be_instance_of(House)
      expect(house.address).to eq('26th and Guerrero')
    end

    it 'returns nil if no associated object' do
      stray_cat = Cat.find(5)
      expect(stray_cat.human).to eq(nil)
    end
  end

  describe '#has_many' do
    let(:ned) { Human.find(3) }
    let(:ned_house) { House.find(2) }

    it 'fetches `cats` from `Human`' do
      expect(ned).to respond_to(:cats)
      cats = ned.cats

      expect(cats.length).to eq(2)

      expected_cat_names = %w(Haskell Markov)
      2.times do |i|
        cat = cats[i]

        expect(cat).to be_instance_of(Cat)
        expect(cat.name).to eq(expected_cat_names[i])
      end
    end

    it 'fetches `humans` from `House`' do
      expect(ned_house).to respond_to(:humans)
      humans = ned_house.humans

      expect(humans.length).to eq(1)
      expect(humans[0]).to be_instance_of(Human)
      expect(humans[0].fname).to eq('Ned')
    end

    it 'returns an empty array if no associated items' do
      catless_human = Human.find(4)
      expect(catless_human.cats).to eq([])
    end
  end

  describe '::assoc_options' do
    it 'defaults to empty hash' do
      class TempClass < DynamicArchive::Base
      end

      expect(TempClass.assoc_options).to eq({})
    end

    it 'stores `belongs_to` options' do
      cat_assoc_options = Cat.assoc_options
      human_options = cat_assoc_options[:human]

      expect(human_options).to be_instance_of(DynamicArchive::BelongsToOptions)
      expect(human_options.foreign_key).to eq(:owner_id)
      expect(human_options.class_name).to eq('Human')
      expect(human_options.primary_key).to eq(:id)
    end

    it 'stores options separately for each class' do
      expect(Cat.assoc_options).to have_key(:human)
      expect(Human.assoc_options).to_not have_key(:human)

      expect(Human.assoc_options).to have_key(:house)
      expect(Cat.assoc_options).to_not have_key(:house)
    end
  end

  describe '#has_many_through' do
    before(:all) do
      class House
        has_many_through :cats, :humans, :cats

        self.finalize!
      end

      class Toy
        has_many_through :cats, :play_times, :cat

        self.finalize!
      end

      class Cat
        has_many_through :toys, :play_times, :toy

        self.finalize!
      end
    end

    let(:house) { House.find(1) }
    let(:breakfast) { Cat.find(1) }
    let(:stringy) { Toy.find(2) }

    it 'adds getter method' do
      expect(house).to respond_to(:cats)
    end

    it 'fetches associated `cats` for a `house`' do
      cats = house.cats

      expect(cats).to be_instance_of(Array)
      expect(cats.length).to eq(2)
      expect(cats[0]).to be_instance_of(Cat)

      expect(cats[0].name).to eq('Breakfast')
      expect(cats[1].name).to eq('Earl')
    end

    it 'fetches associated `toys` for a `cat`' do
      toys = breakfast.toys

      expect(toys).to be_instance_of(Array)
      expect(toys.length).to eq(3)
      expect(toys[0]).to be_instance_of(Toy)

      expect(toys[0].name).to eq('fuzzy')
      expect(toys[1].name).to eq('stringy')
    end

    it 'fetches associated `cats` for a `toy`' do
      cats = stringy.cats

      expect(cats).to be_instance_of(Array)
      expect(cats.length).to eq(2)
      expect(cats[0]).to be_instance_of(Cat)

      expect(cats[0].name).to eq('Breakfast')
      expect(cats[1].name).to eq('Earl')
    end
  end

  describe '#has_one_through' do
    before(:all) do
      class Cat
        has_one_through :home, :human, :house

        self.finalize!
      end
    end

    let(:cat) { Cat.find(1) }
    let(:cat2) { Cat.find(4) }

    it 'adds getter method' do
      expect(cat).to respond_to(:home)
    end

    it 'fetches associated `home` for a `Cat`' do
      house = cat.home

      expect(house).to be_instance_of(House)
      expect(house.address).to eq('26th and Guerrero')

      house2 = cat2.home

      expect(house2.address).to eq('Dolores and Market')
    end
  end
end
