require 'thor'
require 'active_support'
require 'active_support/core_ext'
require 'active_support/inflector'

$thor_runner ||= false

module Trails
  class Generate < Thor
    desc "migration <name>", "adds a migration file to 'db/migrations'"
    def migration(name)
      file_name = name
      timestamp = Time.now.to_i

      path = File.join(PROJECT_ROOT, "db/migrations/#{timestamp}_#{file_name}.sql")
      raise "Migration already created" if File.exist?(path)

      File.new(path, "w")
      puts "Migration #{file_name} can be found at #{path}"
    end

    desc "model <name>", "creates a model template file and a migration file"
    def model(name)
      model_path = File.join(PROJECT_ROOT, "app/models/#{name.underscore}.rb")
      raise "Model already created" if File.exist?(model_path)

      File.open(model_path, "w") do |file|
        file.write("class #{name.camelize} < DynamicArchive::Base\n")
        file.write("  self.finalize!\n")
        file.write("end\n")
      end

      migration(name)
      puts "Model #{name.camelize} can be found at #{model_path}"
    end

    desc "controller <name>", "creates a controller file and a view folder"
    def controller(name)
      file_name = name.pluralize.underscore
      controller_name = "#{name.pluralize.camelize}Controller"
      controller_path = File.join(PROJECT_ROOT, "app/controllers/#{file_name}_controller.rb")
      view_dir = File.join(PROJECT_ROOT, "app/views/#{file_name}")
      raise "Controller already created" if File.exist?(controller_path)

      File.open(controller_path, "w") do |file|
        file.write("class #{controller_name} < ApplicationController\n")
        file.write("end\n")
      end
      Dir.mkdir(view_dir) unless Dir.exist?(view_dir)

      puts "Controller #{controller_name} can be found at #{controller_path}"
    end
  end

  class DB < Thor
    desc "reset", "clears the database, runs all migrations, and re-seeds"
    def reset
      DBConnection.reset
    end

    desc "seed", "seeds the database"
    def seed
      DBConnection.seed
    end

    desc "migrate", "runs all pending migrations"
    def migrate
      DBConnection.migrate
    end
  end

  class CLI < Thor
    desc "generate COMMANDS ...ARGS", "generate new model, migration, or controller"
    subcommand "generate", Generate

    desc "g COMMANDS ...ARGS", "generate new model, migration, or controller"
    subcommand "g", Generate

    desc "db COMMANDS", "alter the database"
    subcommand "db", DB

    desc "server", "starts the trails server, default port: 3000"
    option :port, :default => 3000
    option :host, :default => "localhost"
    def server
      raise "Cannot find Trails App" unless PROJECT_ROOT
      require_relative "#{PROJECT_ROOT}/config/routes"
      require_relative '../bin/server'

      server = TrailsServer.new(options[:port], options[:host])
      server.start
    end

    desc "s", "starts the trails server, default port: 3000"
    option :port, :default => 3000
    option :host, :default => "localhost"
    alias_method :s, :server

    desc "new NAME", "creates a new Trails application skeleton in '/<name>'"
    def new(name)
      app_destination = File.join(Dir.pwd, name)
      raise "Directory already exists at #{app_destination}" if Dir.exist?(app_destination)

      FileUtils.copy_entry(TEMPLATE_PATH, app_destination)
      Dir.mkdir(File.join(app_destination, "app/models"))
      Dir.mkdir(File.join(app_destination, "db/migrations"))
    end
  end
end
